/* 
   HTTP-DAV client routines.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This file is a collection of routines to implement a basic WebDAV
 * client, including an HTTP/1.1 client. */

/* FIXME: Ugly: This is only compiled in if USE_DAV is defined.
 * autoconf might make this nicer.
 * TODO:
 * String overflow checking.
 * Better error condition handling in http_request.
 * Handle Authentication-Info: response header... find a server that sends it.
 */

/* How it works....
 * It's all done by http_request(), basically.
 */

#ifdef USE_DAV

#include <sys/types.h>

#include <sys/stat.h>
#include <sys/time.h>
#include <netinet/in.h>

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "protocol.h"
#include "httpdav.h"
#include "common.h"
#include "socket.h"
#include "md5.h"

/**** Public Interface ****/

/* Connection information... */
int http_sock;
/*bool http_persistent;  Whether persistent connections work */
/* For persistent connections, this is set to false when the socket
 * is closed by the server, so the next request has to open it again.
 * For non-persistent connections, this is not used */
bool http_connected;

/* Time in seconds to wait for the server to give us a 100 Continue
 * after submitting a PUT with an "Expect: 100-continue" header.
 * This doesn't really matter too much */
#define CONT_TIMEOUT 15
int http_expect_works; /* == 0 if it we don't know, 1 if it does,
			    -1 if it doesn't support 100-continue */

/**** Private Details ***/

/* Handy macro to free things */
#define DOFREE(x) if( x!=NULL ) free( x )

#define EOL "\r\n"

char *http_useragent = TARGET "/" VERSION;

/* We need to remember the remote host and port even after connecting
 * the socket so we can form the Destination: header */
char *http_remotehost;
struct in_addr http_remoteaddr;
int http_remoteport;

/* HTTP Authorization. 
 *   http_auth holds the type of authorization last used.
 *   http_auth_cookie holds the last cookie used for authorization:
 *      - for basic auth, this is the Base64 of username:password
 *      - for digest auth, this is... something else.
 *   http_username and http_password hold the username and password
 *     last used.
 */
enum http_auth_t http_auth = http_auth_none;
char *http_username, *http_password;
/* bool http_can_auth; true if we can supply auth details */
bool http_auth_success;
char *http_auth_cookie;
char *http_auth_realm, *http_auth_opaque, *http_auth_nonce;

char *http_status_line;

struct proto_driver dav_driver = {
    http_connect, /* Connect */
    http_userpass, /* Login */
    http_disconnect, /* Disconnect */
    dav_move, /* File move */
    http_put, /* File upload - HTTP PUT */
    http_delete, /* File delete */
    NULL, /* File chmod */
    dav_mkcol, /* Dir create */
    http_delete, /* Dir remove, same as file remove */
    &http_status_line,
    "http", /* Service name */
    "HTTP-DAV" /* User-visible protocol name */
};

/* This doesn't really connect to the server.
 * Returns
 *  PROTO_LOOKUP if the hostname of the server could not be resolved
 *  PROTO_LOCALHOST if the hostname of the local machine could not be
 *    found
 *  PROTO_CONNECT if the socket could not be connected
 *  PROTO_OK if the connection was made successfully.
 */
int http_connect( const char *hostname, const int port ) {
    /* Look up the remote host name */
    if( http_remotehost!=NULL ) free( http_remotehost );
    http_remotehost = strdup( hostname );
    http_remoteport = port;
    if( host_lookup( http_remotehost, &http_remoteaddr ) )
	return PROTO_LOOKUP;
    http_connected = false;
    http_expect_works = 0; /* we don't know yet */
    /* Initialize the authentication variables */
    DOFREE( http_auth_nonce );
    DOFREE( http_auth_realm );
    DOFREE( http_auth_opaque );
    http_auth = http_auth_none;
    return PROTO_OK;
}

/* Supply the user-specified login name and password */
int http_userpass( const char *username, const char *password ) {
    DOFREE( http_username );
    DOFREE( http_password );
    http_username = strdup( username );
    http_password = strdup( password );
    return PROTO_OK;
}

/* Disconnect from the server */
int http_disconnect( void ) {
    if( http_connected ) {
	http_close( );
    }
    return PROTO_OK;
}

/* This parses the response status line, and plonks the numeric
 * status value in status.
 * Returns: PROTO_OK on success, PROTO_ERROR otherwise */
int http_parse_status( char *status_line, int *status ) {
    char *part;
    debug( debug_some, "HTTP response line: %s", status_line );
    /* Save the line for error codes later */
    DOFREE( http_status_line );
    http_status_line = strdup( status_line );
    /* Strip off the CRLF for the status line */
    if( (part = strchr( http_status_line, '\r' )) != NULL )
	*part = '\0';
    /* Parse the status code */
    part = strtok( status_line, " " ); /* "HTTP/1.1" */
    if( part==NULL ) return PROTO_ERROR;
    part = strtok( NULL, " " ); /* xxx */
    if( part==NULL ) return PROTO_ERROR;
    *status = atoi( part );
    debug( debug_some, "HTTP reponse code: %d\n", *status );
    return PROTO_OK;
}

/* Add the authorization headers to the request. 
 * FIXME: This is getting to the stage of ungodly amounts of
 * parameters being passed. Maybe integrate it in a whopper http_request
 * function? Better, have some kind of 'request' struct which has
 * method, uri, date etc.
 * FIXME: This probably leaks like a sieve.
 */
void http_auth_request( char *headers, const char *request_method,
			const char *request_uri, FILE *req_body,
			struct stat *req_body_st, char *date ) {
    char tmp[BUFSIZ];
    char *a1, *md5_a1, *a2, *x, *y, *z, *temp, *auth_response,
	*v, *w, *ei; /* yarg, that enough? */

    switch( http_auth ) {
    case http_auth_basic:
	strcat( headers, "Authorization: Basic " );
	strcat( headers, http_auth_cookie );
	strcat( headers, EOL );
	break;
    case http_auth_digest:
/* RFC2069 Says:
--- begins ---
     response-digest     =
          <"> < KD ( H(A1), unquoted nonce-value ":" H(A2) > <">

     A1             = unquoted username-value ":" unquoted realm-value
                                                ":" password
     password       = < user's password >
     A2             = Method ":" digest-uri-value
--- ends ---

Which means, in English...
   A1 = username + ":" + realm + ":" password
   A2 = method + ":" + uri
   x = nonce + ":" + MD5( A2 )
   y = MD5( A1 ) + ":" + x
   z = MD5( y )
   ( and response-digest = "\"" + z + "\"", but we only need to 
   calculate z, which is put into http_auth_response );
*/
	a1 = malloc( strlen( http_username ) +
		     strlen( http_auth_realm ) +
		     strlen( http_password ) + 3 );
	strcpy( a1, http_username );
	strcat( a1, ":" );
	strcat( a1, http_auth_realm );
	strcat( a1, ":" );
	strcat( a1, http_password );
	debug( debug_total, "A1: %s\n", a1 );
	a2 = malloc( strlen( request_method ) +
		     strlen( request_uri ) + 2 );
	strcpy( a2, request_method );
	strcat( a2, ":" );
	strcat( a2, request_uri );
	debug( debug_total, "A2: %s\n", a2 );
	temp = md5_hex( a2, strlen( a2 ) );
	free( a2 );
	x = malloc( strlen( http_auth_nonce ) +
		    strlen( temp ) + 2 );
	strcpy( x, http_auth_nonce );
	strcat( x, ":" );
	strcat( x, temp );
	free( temp );
	debug( debug_total, "x: %s\n", x );
	md5_a1 = md5_hex( a1, strlen( a1 ) );
	free( a1 );
	y = malloc( strlen( md5_a1 ) +
		    strlen( x ) + 2 );
	strcpy( y, md5_a1 ); 
	strcat( y, ":" );
	strcat( y, x );
	/* take a DEEP breath */
	auth_response = md5_hex( y, strlen( y ) );
	debug( debug_some, "Response digest: %s\n", temp );
	free( x );
	free( y );
	snprintf( tmp, BUFSIZ, "Authorization: Digest "
		  "username=\"%s\", realm=\"%s\", "
		  "nonce=\"%s\", uri=\"%s\", response=\"%s\"",
		  http_username,
		  http_auth_realm,
		  http_auth_nonce,
		  request_uri,
		  auth_response );

	free( auth_response );

/* Now for the entity digest. RFC2069 says:
--- begins ---
entity-digest = <"> KD (H(A1), unquoted nonce-value ":" Method ":"
                           date ":" entity-info ":" H(entity-body)) <">
       ; format is <"> *LHEX <">

date = = rfc1123-date            ; see section 3.3.1 of [2]
entity-info = H(
          digest-uri-value ":"
          media-type ":"         ; Content-type, see section 3.7 of [2]
          *DIGIT ":"             ; Content length, see 10.12 of [2]
          content-coding ":"     ; Content-encoding, see 3.5 of [2]
          last-modified ":"      ; last modified date, see 10.25 of [2]
          expires                ; expiration date; see 10.19 of [2]
          )

last-modified   = rfc1123-date  ; see section 3.3.1 of [2]
expires         = rfc1123-date
--- ends ---

Since we do not pass last-modified, content-encoding, expires, or
content-encoding headers, we can skip those fields.

So what we do is this:

   ei = "::" + req_body_st->st_size + ":::"
   y = MD5( req_body )
   date = current date (in format specified by RFC1123)
   z = nonce + ":" + method + ":" + date + ":" + ei + ":" + y
   md5_a1 = MD5( A1 )    (from above)
   v = md5_a1 + ":" + z
   w = MD5( v )   where w is the unquoted entity-digest
*/
	if( req_body!=NULL ) {
	    temp = malloc( BUFSIZ );
	    snprintf( temp, BUFSIZ, "%d", req_body_st->st_size );
	    ei = malloc( strlen(temp) + 5 + 1 );
	    strcpy( ei, "::" );
	    strcat( ei, temp ); free( temp );
	    strcat( ei, ":::" );
	    debug( debug_total, "Entity-info: %s\n", ei );
	    y = md5_hex_stream( req_body );
	    rewind( req_body ); /* leave it at the beginning */
	    z = malloc( strlen( http_auth_nonce ) + strlen( request_method ) +
			strlen( date ) + strlen( ei ) + strlen( y ) + 5 );
	    strcpy( z, http_auth_nonce );
	    strcat( z, ":" );
	    strcat( z, request_method );
	    strcat( z, ":" );
	    strcat( z, date );
	    strcat( z, ":" );
	    strcat( z, ei ); free( ei );
	    strcat( z, ":" );
	    strcat( z, y ); free( y );
	    debug( debug_total, "Z: %s\n", z );
	    v = malloc( strlen( md5_a1 ) + strlen( z ) + 2 );
	    strcpy( v, md5_a1 );
	    strcat( v, ":" );
	    strcat( v, z ); free( z );
	    debug( debug_total, "V: %s\n", v );
	    w = md5_hex( v, strlen( v ) );
	    free( v );
	    debug( debug_total, "W: %s\n", w );
	    /* Now lob it on the end */
	    strcat( tmp, ", digest=\"" );
	    strcat( tmp, w ); free( w );
	    strcat( tmp, "\"" );
	} else {
	    free( md5_a1 );
	}

	if( http_auth_opaque!=NULL ) {
	    strcat( tmp, ", opaque=\"" );
	    strcat( tmp, http_auth_opaque );
	    strcat( tmp, "\"" );
	}
	strcat( tmp, EOL );
	strcat( headers, tmp );
	break;
    default:
	break;
    }
}

/* This is called when a WWW-Authenticate header is received in the response.
 * It is used to perform any necessary computation required for presenting
 * authorisation information on the next (retry-with-auth) request.
 * i.e., to work out http_auth_cookie.
 * Basic auth is simple - we just take the string
 *   username:password
 * and Base64 it.
 * Digest auth is complex - read the RFC.
 */
void http_auth_challenge( char *challenge ) {
    /* Parse the authentication challenge */
    char auth_type[BUFSIZ], auth_parms[BUFSIZ];
    char *part, *temp, *key, *val, *keypnt;
    size_t len, len2;
    /* FIXME: Decent parser.
     * Here, we split the entire challenge into the challenge part and the
     * other parameters - everything before, and after the first space char,
     * respectively. 
     */
    len = strcspn( challenge, " " );
    debug( debug_total, "Challenge: [%s]\n", challenge, len );
    if( len == 0 ) {
	debug( debug_some, "Invalid challenge header: [%s]\n", challenge );
	return;
    }
    strncpy( auth_type, challenge, len );
    *(auth_type+len) = '\0'; /* null-term */
    strcpy( auth_parms, challenge+len+1 );
    /* Now process it */
    debug( debug_total, "Challenge: type [%s] parms [%s]\n", auth_type,
	   auth_parms );
    if( strcasecmp( auth_type, "basic" ) == 0 ) {
	if( http_auth==http_auth_digest ) {
	    /* We're being spoofed, don't hand out the password */
	    debug( 0, 
"WARNING: Spoof attack probable. The server has switched to using basic\n"
"authentication from digest authenticaion, which may be an attempt to\n"
"discover your password. Basic auth will NOT be used.\n" );
	    return;
	}
	debug( debug_some, "Using basic authentication.\n" );
	http_auth = http_auth_basic;
	temp = malloc( strlen(http_username) + strlen(http_password) + 2 );
	strcpy( temp, http_username );
	strcat( temp, ":" );
	strcat( temp, http_password );
	http_auth_cookie = base64( temp );
	debug( debug_full, "Basic auth cookie: %s\n", http_auth_cookie );
	http_auth = http_auth_basic;
	free( temp );
    } else if( strcasecmp( auth_type, "digest" ) == 0 ) {
	/* Digest authentication */
	debug( debug_some, "Trying digest authentication.\n" );
	/* FIXME: Another decent parser */
	part = strtok( auth_parms, "," );
	/* Chuck out the old nonce + realm */
	DOFREE( http_auth_nonce );
	DOFREE( http_auth_realm );
	DOFREE( http_auth_opaque );
	while( part!=NULL ) {
	    /* Get the first bit before the = */
	    len = strcspn( part, "=" );
	    key = malloc( len + 1 );
	    keypnt = key;
/*	    debug( debug_some, "Key len: %d\n", len ); */
	    strncpy( key, part, len );
	    *(key+len) = '\0'; /* null-terminate */
	    while( *keypnt==' ' ) keypnt++; /* skip leading whitespace */
	    /* Value is ALWAYS a quoted string */
	    len2 = strlen( part ) - (len+3);
/*	    debug( debug_some, "Val len: %d\n", len2 ); */
	    if( len2 > 0 ) {
		val = malloc( len2 + 1 );
		strncpy( val, part+len+2, len2 );
		*(val+len2) = '\0'; /* null-terminate */
		debug( debug_total, "Got: [%s] - [%s]\n", keypnt, val );
		if( strcasecmp( keypnt, "nonce" ) == 0 ) {
		    http_auth_nonce = val;
		} else if( strcasecmp( keypnt, "realm" ) == 0 ) {
		    http_auth_realm = val;
		} else if( strcasecmp( keypnt, "opaque" ) == 0 ) {
		    http_auth_opaque = val;
		} else {
		    /* it's no use, chuck it */
		    free( val );
		}
	    }
	    part = strtok( NULL, "," );
	    free( key );
	}
	debug( debug_full, "Nonce: %s, realm: %s\n",
	       http_auth_nonce, http_auth_realm );    
	if( (http_auth_nonce != NULL) && (http_auth_realm != NULL) ) {
	    http_auth = http_auth_digest;
	} else {
	    http_auth = http_auth_none;
	    debug( debug_some, "Could not use digest authentication.\n" );
	}
    } else {
	http_auth = http_auth_none;
	debug( debug_some, "Unrecognised authentication method.\n" );
    }
}

/* TODO: The request / response mechanism is based on a single function call:
 * i.e.
 *   http_request( "My-Header: blah\r\nAnother-header: bloo\n", ... )
 * Which returns a simple response code (TODO: and body?).
 * This transparently supports authentication.
 * If the request has a body, pass this as req_body, else req_body MUST 
 * be NULL.
 *
 * Returns:
 *   PROTO_OK if the request was made (whether 'successful' or not)
 *   PROTO_ERROR if the request could not be made
 * The STATUS CODE is placed in resp_status.
 * 
 * This is all a bit involved, but it
 */
int http_request( const char *method, 
		  const char *request_uri,
		  const char *extra_headers,
		  int *resp_status, 
		  FILE *req_body ) {
    char req_header[BUFSIZ], tmp[BUFSIZ], buffer[BUFSIZ],
	*part, *name, *value, *date;

    int ret, length, attempt, status;
    fd_set sockfd;
    struct timeval tv;
    struct stat bodyst;
    bool using_expect, /* whether we have sent a Expect: 100 header */
	close_connection;
    enum http_te_t encoding;
    /* Initialization... */
    debug( debug_some, "Request started...\n" );
    DOFREE( http_status_line );
    http_status_line = strdup( "Unknown error.\n" );
    ret = PROTO_OK;

    /* Do extra stuff if we have a body */
    if( req_body!=NULL ) {
	if( fstat( fileno(req_body), &bodyst ) < 0 )
	    /* Stat failed */
	    return PROTO_ERROR;
	snprintf( tmp, BUFSIZ, "Content-Length: %d" EOL, bodyst.st_size );
	strcat( req_header, tmp );
    }

    /* I shall try this only twice... 
     * First time, with default authentication stuff (either, what we
     * did last time, or none at all), then with up-to-the-minute
     * what-the-server-requested authentication stuff. */

    attempt = 1;

    do {

	/* Form the basic request header - this doesn't change */
	snprintf( req_header, BUFSIZ, 
		  "%s %s HTTP/1.1" EOL
		  "Host: %s:%d" EOL
		  "User-Agent: %s" EOL,
		  method, request_uri,
		  http_remotehost, http_remoteport,
		  http_useragent );

	/* Add in the other headers supplied by the calling function */
	if( extra_headers!=NULL )
	    strcat( req_header, extra_headers );

	date = rfc1123_date( time(NULL) );

	if( http_auth!=http_auth_none ) {
	    /* Add the authorization headers in */
	    http_auth_request( req_header, method, request_uri, req_body,
		&bodyst, date );
	}
    
	/* Now handle the body. */
	if( req_body!=NULL ) {
	    /* Add Content-Length and Date headers. */
	    sprintf( tmp, 
		     "Content-Length: %ld" EOL
		     "Date: %s" EOL,
		     bodyst.st_size, date );
	    strcat( req_header, tmp );
	    if( http_expect_works > -1 ) {
		/* Add Expect: 100-continue. */
		strcat( req_header, "Expect: 100-continue" EOL );
		using_expect = true;
	    } else {
		using_expect = false;
	    }	    
	} else {
	    using_expect = false;
	}

	free( date ); /* that's enough of that */

	/* End of headers... */
	strcat( req_header, EOL );
	
	/* Now send the request */

	/* Open the connection if necessary */
	if( !http_connected ) {
	    if( (ret = http_open()) != PROTO_OK )
		return ret;
	}

	debug( debug_some, "Sending request:\n%s", req_header );
	/* Send the headers */
	send_data( http_sock, req_header );
		
	debug( debug_some, "Request sent.\n" );

	/* Now, if we are doing a Expect: 100, hang around for a short
	 * amount of time, to see if the server actually cares about the 
	 * Expect and sends us a 100 Continue response if the request
	 * is valid, else an error code if it's not. This saves sending
	 * big files to the server when they will be rejected. 
	 */
	
	if( using_expect ) {
	    debug( debug_some, "Waiting for response...\n" );
	    /* And wait for something to happen for a while */
	    /* Set up the fd set for the select() */
	    FD_ZERO( &sockfd );
	    FD_SET( http_sock, &sockfd );
	    /* Set the timeout appropriately */
	    tv.tv_sec = CONT_TIMEOUT;
	    tv.tv_usec = 0;
	    ret = select( http_sock+1, &sockfd, NULL, NULL, &tv );
	    
	    switch( ret ) {
	    case -1: /* error */
		debug( debug_some, "Wait went wrong.\n" );
		perror( "select" );
		close_connection = true;
		break;
	    case 1: /* we got us a response! */
		debug( debug_some, "Wait got data.\n" );
		http_expect_works = 1; /* it works! */
		break;
	    case 0: /* timed out */
		debug( debug_some, "Wait timed out.\n" );
		http_expect_works = -1; /* don't try that again */
		/* and give them the body */
		send_stream( http_sock, req_body, bodyst.st_size );
		rewind( req_body ); /* since we may have to send it again */
		break;
	    }
	} else if( req_body!=NULL ) {
	    /* Just chuck the file down the socket */
	    debug( debug_some, "Sending body...\n" );
	    send_stream( http_sock, req_body, bodyst.st_size );
	    rewind( req_body ); /* we may have to send it again */
	    send_line( http_sock, "" );
	    debug( debug_some, "Body sent.\n" );
	}
	
	/* Now, we have either:
	 *   - Sent the header and body, or
	 *   - Sent the header incl. Expect: line, and got some response.
	 * In any case, we get the status line of the response.
	 */

	if( read_line( http_sock, tmp, BUFSIZ ) < 0 ) {
	    /* Couldn't get the status line. Bail out. */
	    close_connection = true;
	    ret = PROTO_ERROR;
	    goto http_request_finish_proc;
	} 

	debug( debug_some, "< %s", tmp );

	/* Got the status line - parse it */
	ret = http_parse_status( tmp, &status );
	
	if( using_expect && (status == 100) ) {
	    /* We are using Expect: 100, and we got a 100-continue 
	     * return code... send the request body */
	    debug( debug_some, "Got continue... sending body now...\n" );
	    /* But first, skip the rest of the headers, we're not
	     * interested in them */
	    do {
		if( read_line( http_sock, tmp, BUFSIZ ) < 0 ) {
		    close_connection = true;
		    ret = PROTO_ERROR;
		    goto http_request_finish_proc;
		} 
	    } while( strcmp( tmp, EOL ) != 0 );
	    send_stream( http_sock, req_body, bodyst.st_size );
	    rewind( req_body );
	    debug( debug_some, "Body sent.\n" );
	    /* Now get the status line again */
	    if( read_line( http_sock, tmp, BUFSIZ ) < 0 ) {
		close_connection = true;
		ret = PROTO_ERROR;
		goto http_request_finish_proc;
	    } 
	    debug( debug_some, "< %s", tmp );

	    /* Got the status line - parse it */
	    ret = http_parse_status( tmp, &status );

	}
	
	/* We've got the status line... now get the headers */
	
	/* Body length is tres important, since if we know it, we can
	 * do persistent connections. 
	 * NO body is returned if the method is HEAD, or the resp status
	 * is 204, 304 or 1xx (HTTP/1.1 draft 5 section 4.3)
	 * length==-1 if we do NOT know the body length.
	 * length>-1 if we DO know the body length.
	 */
	if( (strcmp( method, "HEAD" ) == 0 ) ||
	    status==204 ||
	    status==304 ||
	    (status/100 == 1) ) {
	    length = 0;
	} else {
	    length = -1;
	}
	encoding = http_te_unknown;
	close_connection = false;
	
	/* Now read the rest of the header... up to the next blank line */
	while( read_line( http_sock, buffer, BUFSIZ ) > 0 ) {
	    /* FIXME: What to check for here? */
	    if( strcmp( buffer, EOL ) == 0 ) {
		debug( debug_total, "End of headers.\n" );
		break;
	    }
	    debug( debug_some, "< [%d] %s", strlen( buffer ), buffer );
	    /* Now parse the header line. This is all a bit noddy. */
	    name = strtok( buffer, ":" );
	    if( name != NULL ) {
		value = strtok( NULL, ":" );
		/* Strip leading whitespace, and trailing CRLF */
		while( *value!='\0' && *value==' ' ) value++;
		if( (part = strchr( value, '\r' )) != NULL ) 
		    *part = '\0';
		if( (part = strchr( value, '\n' )) != NULL ) 
		    *part = '\0';
		debug( debug_some, "Header: [%s] = [%s]\n", name, value );
		if( strcasecmp( name, "Content-Length" ) == 0 ) {
		    length = atoi( value );
		} else if( strcasecmp( name, "Transfer-Encoding" ) == 0 ) {
		    if( strcasecmp( value, "chunked" ) == 0 ) {
			encoding = http_te_chunked;
		    } else {
			encoding = http_te_unknown;
		    }
		} else if( strcasecmp( name, "Connection" ) == 0 ) {
		    if( strcasecmp( value, "close" ) == 0 ) {
			close_connection = true;
		    }
		} else if( strcasecmp( name, "WWW-Authenticate" ) == 0 ) {
		    /* Parse the authentication challenge */
		    http_auth_challenge( value );
		}
	    }
	}
	if( length==-1 && encoding==http_te_unknown ) {
	    /* Then we don't know how to get the body. Oh dear, give up. */
	    debug( debug_some, "Response body is lost on me.\n" );
	    /* And the connection will have to be closed */
	    close_connection = true;
	} else if( encoding==http_te_chunked ) {
	    /* The response body is encoded using 'chunks'.
	     * HTTP/1.1 draft 05 tells how to do this.
	     * It goes:  `SIZE CRLF CHUNK CRLF SIZE CRLF CHUNK CRLF ...'
	     * ended by a `SIZE CHUNK 0 CRLF', 0-byte chunk */
	    
	    int chunk_size; /* the size of the next chunk */
	    char *chunk; /* the data of the next chunk */
	    
	    do {
		if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
		    close_connection = true;
		    ret = PROTO_ERROR;
		    goto http_request_finish_proc;
		}
		sscanf( buffer, "%x", &chunk_size );
		debug( debug_full, "Got chunk size: %d\n", chunk_size );
		if( chunk_size > 0 ) {
		    chunk = malloc( chunk_size + 1 );
		    if( read_data( http_sock, chunk, chunk_size ) < 0 ) {
			/* It broke */
			close_connection = true;
			free( chunk );
			goto http_request_finish_proc;
		    }
		    *(chunk+chunk_size) = '\0';
		    debug( debug_full, "<<< Chunk:\n%s\n", chunk );
		    free( chunk ); /* chuck the chunk, we don't want it */
		    /* Skip CRLF */
		    if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
			/* It broke */
			close_connection = true;
			ret = PROTO_ERROR;
			goto http_request_finish_proc;
		    }
		}
	    } while( chunk_size > 0 ); /* loop until 0-length chunk */
	    /* Read in that last line */
	    if( read_line( http_sock, buffer, BUFSIZ ) < 0 ) {
		/* It broke, kill it */
		close_connection = true;
		goto http_request_finish_proc;
	    }
	} else if( length > 0 ) {
	    int block_size;
	    /* No encoding, but we gotta have a Content-Length.
	     * FIXME: Erm, will this work? */
	    while( length > 0 ) {
		if( (block_size = read( http_sock, buffer, BUFSIZ )) < 0 ) {
		    /* It broke */
		    close_connection = true;
		    goto http_request_finish_proc;
		}
		length -= block_size;
	    }
	} else {
	    /* We don't know how long the body is, so give up */
	    close_connection = true;
	}
    
http_request_finish_proc:
	/* Now, do we close the connection? */
	if( close_connection ) {
	    http_close( );
	}
    
	if( http_auth == http_auth_none )
	    /* We can't authenticate... give up.
	     * http_auth will have been set by http_auth_challenge
	     * by know if it found out a way in which we can 
	     * authenticate with the server. */
	    break;
	
	/* Now, do that all *again* if it didn't work.
	 * Otherwise, give up */

    } while( (++attempt<3) && (status==401) );
    
    *resp_status = status;
    
    debug( debug_some, "Request ends. Status: %s\n", http_status_line );

    return ret;
}

/* Simple HTTP put. 
 * local is the local filename. Remote is the destination URI (URI?)
 * Make it proper.
 * Returns:
 *   PROTO_FILE if no local file
 *   PROTO_ERROR if something general goes wrong
 *   PROTO_OK if it all works fine
 */
int http_put( const char *local, const char *remote ) {
    int status, ret;
    FILE *f;
    
    if( (f = fopen( local, "r" )) == NULL )
	return PROTO_ERROR;

    ret = http_request( "PUT", remote, NULL, &status, f );

    fclose( f );

    if( ret!=PROTO_OK ) return PROTO_ERROR;

    if( status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

/* Perform the file operations */
int dav_move( const char *from, const char *to ) {
    char headers[BUFSIZ];
    int status;

    snprintf( headers, BUFSIZ, "Destination: http://%s:%d%s" EOL, 
	      http_remotehost, http_remoteport, to );

    if( http_request( "MOVE", from, headers, &status, NULL ) != PROTO_OK )
	return PROTO_ERROR;	
    
    if( status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

/* Deletes the specified resource on the server */
int http_delete( const char *filename ) {
    int status;
    
    if( http_request( "DELETE", filename, NULL, &status, NULL ) != PROTO_OK )
	return PROTO_ERROR;

    if( status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

int dav_mkcol( const char *dirname ) {
    int status;

    if( http_request( "MKCOL", dirname, NULL, &status, NULL ) != PROTO_OK )
	return PROTO_ERROR;

    if( status/100 == 2 ) {
	return PROTO_OK;
    } else {
	return PROTO_ERROR;
    }

}

int http_open( void ) {
    debug( debug_some, "Connecting to %s:%d...\n", http_remotehost, http_remoteport );
    if( (http_sock = socket_connect( http_remoteaddr, http_remoteport )) < 0 )
	return PROTO_CONNECT;
    debug( debug_some, "Connected.\n" );
    http_connected = true;
    return PROTO_OK;
}

int http_close( void ) {
    debug( debug_full, "Closing socket.\n" );
    socket_close( http_sock );
    http_connected = false;
    debug( debug_some, "Socket closed.\n" );
    return PROTO_OK;
}    

/*    int (*file_chmod) ( const char *filename, const mode_t mode );
*/

#endif /* USE_DAV */
