/* 
   sitecopy, remote web site copy manager.
   Copyright (C) 1998, Joe Orton <joe@orton.demon.co.uk>
                                                                     
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
  
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* This is the core functionality of sitecopy, performing updates
 * and checking files etc. */

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <fnmatch.h>

#include "sites.h"
#include "protocol.h"
#include "socket.h"
#include "common.h"
#include "frontend.h"

/* for the protocol driver structures */
extern struct proto_driver ftp_driver;

#ifdef USE_DAV
extern struct proto_driver dav_driver;
#endif

/* This is the maximum number of directories which can be held in the queue 
 * at one time - each entry is only a pointer, so the size is not too
 * much of a problem */
#define MAXDIRS 4000

/* This is the string used in the storage files to indicate the
 * line is a directory rather than a file */
#define DIRWORD "dir"

/* Groovy macro to remove items from a doubly linked list...
 * give it the item to delete and the list head pointer */
#define LL_DELETE(item,head)			\
if( item->prev ) {				\
    item->prev->next = item->next;		\
} else {					\
    head = item->next;				\
}						\
if( item->next ) {				\
    item->next->prev = item->prev;		\
}

/* From sitecopy.c */
extern enum debug_types debug_level;
extern char *progname;
extern char *copypath;

extern bool prompting;

/* This holds ALL the sites defined in the rcfile */
struct site_t *all_sites;
 
/* Functions for dealing with sites */

/* site_readsites will read ~/.sitecopyrc and fill the sites list */
int site_readsites( char *filename ) {
    FILE *fp;
    int state, last_state=8, ret=0;
    bool alpha, hash;
    char buf[BUFSIZ];
    char *ch, *ptr, key[BUFSIZ], val[BUFSIZ], val2[BUFSIZ];
    /* Holders for the site info */
    struct site_t *this_site, *last_site;
    struct name_mapping_t *new;

    if( (fp = fopen( filename, "r" )) == NULL ) {
	perror("fopen");
	return 1;
    } 

    last_site = this_site = NULL;

    while( fgets( buf, sizeof(buf), fp ) != NULL ) {
	state = 0;
	ptr = key;
	memset( key, 0, BUFSIZ );
	memset( val, 0, BUFSIZ );
	memset( val2, 0, BUFSIZ );
	for( ch=buf; *ch!='\0'; ch++ ) {
	    alpha = !isspace(*ch);
	    hash = (*ch == '#');
	    switch( state ) {
	    case 0: /* whitespace at beginning of line */
		if(hash) {
		    state = 8;
		} else if(alpha) {
		    *(ptr++) = *ch;
		    state = 1;
		}
		break;
	    case 1: /* key */
		if(hash) {
		    state = 8;
		} else if(!alpha) {
		    ptr = val;
		    state = 2;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 2: /* whitespace after key */
		if(hash) {
		    state = 8;
		} else if( *ch == '"' ) {
		    state = 4; /* begin quoted value */
		} else if ( alpha ) {
		    *(ptr++) = *ch;
		    state = 3;
		} 
		break;
	    case 3: /* unquoted value 1 */
		if(hash) {
		    state = 8;
		} else if( !alpha ) {
		    ptr = val2;
		    state = 5;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 4: /* quoted value 1 */
		if( *ch == '"' ) {
		    ptr = val2;
		    state = 5;
		} else if( *ch == '\\' ) {
		    last_state = 4;
		    state = 9;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 5: /* whitespace after value 1 */
		if(hash) {
		    state = 8;
		} else if( *ch == '"' ) {
		    state = 6; /* begin quoted value 2 */
		} else if ( alpha ) {
		    *(ptr++) = *ch;
		    state = 7; /* begin unquoted value 2 */
		} 
		break;
	    case 6: /* quoted value 2 */
		if( *ch == '"' ) {
		    state = 8;
		} else if( *ch == '\\' ) {
		    last_state = 4;
		    state = 9;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 7: /* unquoted value 2 */
		if(hash) {
		    state = 8;
		} else if( !alpha ) {
		    state = 8;
		} else {
		    *(ptr++) = *ch;
		}
		break;
	    case 8: /* ignore till end of line */
		break;
	    case 9: /* a literal (\-slashed) in a value */
		*(ptr++) = *ch;
		state = last_state;
		break;
	    }
	}
	
	debug( debug_total, "Key [%s] Value: [%s] Value2: [%s]\n", key, val, val2);
	if( strlen( key ) > 0 ) {
	    if( strcmp( key, "site" ) == 0 ) {
		/* Beginning of a new Site */
		last_site = this_site;
		/* Allocate new item */
		this_site = malloc( sizeof( struct site_t ) );
		/* Zero it out */
		memset( this_site, 0, sizeof( struct site_t ) );
		if( last_site != NULL ) { /* next site */
		    last_site->next = this_site;
		} else { /* First site */
		    all_sites = this_site;
		}
		this_site->name = strdup( val );
		this_site->files = NULL;
		/* defaults... */
		this_site->perms = sitep_ignore;
		this_site->symlinks = sitesym_follow;
		this_site->nodelete = false;
		this_site->checkmoved = false;
		this_site->protocol = proto_ftp;
		this_site->driver = &ftp_driver;
		this_site->ftp_pasv_mode = true;
		/* Now work out the info filename */
		this_site->infofile = malloc( strlen(copypath) + strlen(val) + 1);
		strcpy( this_site->infofile, copypath );
		strcat( this_site->infofile, val );
	    } else if( this_site == NULL ) {
		/* TODO: Only do this if we have a valid keyword */
		ret = SITE_RCCORRUPT;
		break; 
	    } else if( strcmp( key, "username" ) == 0 ) {
		/* username */
		this_site->username = strdup( val );
	    } else if( strcmp( key, "server" ) == 0 ) {
		if( val ) this_site->server = strdup( val );
	    } else if( strcmp( key, "port" ) == 0 ) {
		if( val ) this_site->port = atoi( val );
	    } else if( strcmp( key, "password" ) == 0 ) {
		if( val ) this_site->password = strdup( val );
	    } else if( strcmp( key, "remote" ) == 0 ) {
		if( strlen(val) > 0 ) {
		    /* Relative filenames must start with "~/" */
		    if( val[0] == '~' ) {
			if( strlen(val) > 1 ) {
			    if( val[1] == '/' ) {
				if( val[strlen(val)-1] != '/' )
				    strcat( val, "/" );
				this_site->remote_root = strdup( val + 2 );
			    } else {
				ret = SITE_RCBADTILDE;
				break; 
			    }
			} else {
			    ret = SITE_RCBADREMOTE;
			    break;
			}
		    } else { /* Dirname doesn't begin with a ~ */
			if( val[strlen(val)-1] != '/' )
			    strcat( val, "/" );
			this_site->remote_root = strdup( val );
		    }
		} else { /* strlen(val) == 0 */
		    ret = SITE_RCBADREMOTE;
		    break;
		}
	    } else if( strcmp( key, "local" ) == 0 ) {
		if( val[strlen(val)-1] != '/' )
		    strcat( val, "/" );
		this_site->local_root = strdup( val );
	    } else if( strcmp( key, "permissions") == 0 ) {
		if( strcmp( val, "ignore" ) == 0 ) {
		    this_site->perms = sitep_ignore;
		} else if( strcmp( val, "exec" ) == 0 ) {
		    this_site->perms = sitep_exec;
		} else if( strcmp( val, "all" ) == 0 ) {
		    this_site->perms = sitep_all;
		} else {
		    ret = SITE_RCBADPERMS;
		    break;
		}
	    } else if( strcmp( key, "symlinks" ) == 0 ) {
		if( strcmp( val, "follow" ) == 0 ) {
		    this_site->symlinks = sitesym_follow;
		} 
#if 0 /* not yet - can we do this with FTP? */
		else if( strcmp( val, "maintain" ) == 0 ) {
		    this_site->symlinks = sitesym_maintain;
		}
#endif
		else if( strcmp( val, "ignore" ) == 0 ) {
		    this_site->symlinks = sitesym_ignore;
		} 
	    } else if( strcmp( key, "nodelete" ) == 0 ) {
		this_site->nodelete = true;
	    } else if( strcmp( key, "checkmoved" ) == 0 ) {
		this_site->checkmoved = true;
	    } else if( strcmp( key, "exclude" ) == 0 ) {
		/* Find the position in the array */
		this_site->excludes[this_site->numexcludes] = strdup( val );
		this_site->numexcludes++;
	    } else if( strcmp( key, "protocol" ) == 0 ) {
		if( strcmp( val, "ftp" ) == 0 ) {
		    this_site->protocol = proto_ftp;
		    this_site->driver = &ftp_driver;
		}
#ifdef USE_DAV
		 else if( strcmp( val, "http" ) == 0 ) {
		    this_site->protocol = proto_http;
		    this_site->driver = &dav_driver;
		 }
#endif /* USE_DAV */
	        else {
		    ret = SITE_RCBADPROTOCOL;
		    break;
		}
	    } else if( strcmp( key, "rename" ) == 0 ) {
		if( val==NULL || val2==NULL ) {
		    ret = SITE_RCNOVALUE;
		    break;
		}
		new = malloc( sizeof(struct name_mapping_t) );
		new->next = this_site->renames;
		this_site->renames = new;
		new->local = strdup( val );
		new->remote = strdup( val2 );
	    } else if( strcmp( key, "ftp" ) == 0 ) {
		if( strcmp( key, "nopasv" ) == 0 ) {
		    this_site->ftp_pasv_mode = false;
		}
	    } else {
		ret = SITE_RCBADKEYWORD;
		break; 
	    }
	}
    }

    fclose( fp );
    return ret;
}

int site_find( char *sitename, struct site_t **the_site ) {
    struct site_t *current;

    if( all_sites == NULL) return -1; /* be sane */
    for( current = all_sites; current!=NULL; current=current->next ) {
	if( strcmp( current->name, sitename ) == 0 ) {
	    /* We found it */
	    *the_site = current;
	    return site_verify( current );
	}
    }
    return SITE_NOSITE;
}

/* Returns zero if site is properly defined, else non-zero */
int site_verify( struct site_t *any_site ) {
    struct stat localst;
    char *temp;
    int ret;

    /* Check they specified everything in the rcfile */

    if( any_site->name == NULL ) {
	return SITE_NONAME;
    } else if( any_site->server == NULL ) {
	return SITE_NOSERVER;
    } else if( any_site->username == NULL ) {
	return SITE_NOUSERNAME;
    } else if( any_site->password == NULL ) {
	return SITE_NOPASSWORD;
    } else if( any_site->remote_root == NULL ) {
	return SITE_NOREMOTEDIR;
    } else if( any_site->local_root == NULL ) {
	return SITE_NOLOCALDIR;
    }
    /* Now check the local directory actually exists.
     * To do this, stat `/the/local/root/.', which will fail if the
     * can't read the directory or if it's a file not a directory */
    temp = malloc( strlen( any_site->local_root ) + 2 );
    strcpy( temp, any_site->local_root ); /* must have trailing / */
    strcat( temp, "." ); 
    ret = stat( temp, &localst );
    free( temp );
    if( ret != 0 ) {
	return SITE_ACCESSLOCALDIR;
    }

    /* Assign default ports if they didn't bother to */
    if( any_site->port == 0 ) {
	debug( debug_some, "No port given... finding default.\n" );
	any_site->port = get_tcp_port( any_site->driver->service_name );
	if( any_site->port == 0 ) {
	    ret = SITE_RCBADPORT;
	} else {
	    debug( debug_some, "Assigned default port: %d\n", any_site->port );
	}
    }
    return 0;
}

/* Updates the remote files list of the given site.
 * Tricky, this - have to do a LIST's, and try and
 * parse the response sensibly */
int site_resync( struct site_t *the_site )  {
    /* TODO: Implement this. */
    site_writefiles( the_site );
    return 0;
}

/* Updates the remote site with the local copy.
 * This is the core functionality of sitecopy.
 * All we do is compare the local files list with the remote files
 * list, and upload any changed or new files.
 * Directories are treated differently - any new ones must be created BEFORE
 * any files are uploaded, and any old ones must be deleted AFTER any files
 * are deleted, since most FTP servers will not allow you to DELE a 
 * non-empty directory.
 */
int site_update( struct site_t *the_site )  {
    int ret = 0;
    struct site_file_t *current, *last;
    char temp[BUFSIZ];
    extern bool ftp_use_passive;
    /* screeech... i don't like this at all */
    ftp_use_passive = the_site->ftp_pasv_mode;    
    
    /* Connect to the server */
    printf( "Connecting to server... " );
    if( (*the_site->driver->connect)( the_site->server, the_site->port ) != PROTO_OK ) {
	printf( "could not connect!\n%s\n", *(the_site->driver->last_error) );
	return 2;
    }
    printf( "connected.\nLogging in... " );
    /* Now log in */
    if( (*the_site->driver->login)( the_site->username, the_site->password ) != PROTO_OK ) {
	printf( "could not log in!:\n%s\n", *(the_site->driver->last_error) );
	(*the_site->driver->close)( );
	return 2;
    }
    printf( "logged in.\n" );
    
    for( current=the_site->files; current!=NULL; current=current->next ) {
	if( current->dir && (current->diff == file_new ) ) {
	    /* New dir! */
	    if( prompting ) {
		snprintf( temp, BUFSIZ, "Create directory: %s", current->rel_local );
		if( !fe_ynprompt( temp ) ) continue;
	    } else {
		snprintf( temp, BUFSIZ, "Creating directory: %s\n", current->rel_local );
		fe_status( temp );
	    }
	    if( (*the_site->driver->dir_create)( current->full_remote ) != PROTO_OK ) {
		snprintf( temp, BUFSIZ, "Error! Could not create directory:\n%s\n", *(the_site->driver->last_error) );
		fe_error( temp );
		ret = 1;
	    } else {
		current->updated = true;
	    }
	}
    }
    
    last = NULL;

    for( current=the_site->files; current!=NULL; current=current->next ) {
	if( !current->dir ) {
	    switch( current->diff ) {
	    case file_new: /* File is new, upload it */
		if( prompting ) {
		    snprintf( temp, BUFSIZ, "Upload new file: %s", current->rel_local );
		    if( !fe_ynprompt( temp ) ) break;
		} else {
		    snprintf( temp, BUFSIZ, "Uploading new file: %s\n", current->rel_local );
		    fe_status( temp );
		}
		if( (*the_site->driver->file_upload)( current->full_local, current->full_remote ) != PROTO_OK ) {
		    snprintf( temp, BUFSIZ, "Error! Could not upload file:\n%s\n", *(the_site->driver->last_error) );
		    fe_error( temp );
		    ret = 1;
		} else { 
		    /* Successful upload! */
		    current->updated = true;
		    if( (the_site->perms == sitep_all) ||
			( (current->mode & S_IXUSR) &&
			  (the_site->perms == sitep_exec) ) ) {
			/* We need to chmod the file ... */
			if( prompting ) {
			    snprintf( temp, BUFSIZ, "Set file permissions to %03o", current->mode & 0777 );
			    if( !fe_ynprompt( temp ) ) break;
			} else {
			    snprintf( temp, BUFSIZ, "Setting file permissions to %03o\n", current->mode & 0777);
			    fe_status( temp );
			}
			if( (*the_site->driver->file_chmod)( current->full_remote, current->mode ) != PROTO_OK ) {
			    snprintf( temp, BUFSIZ, "Error! Could not set permissions:\n%s\n", *(the_site->driver->last_error) );
			    fe_error( temp );
			    ret = 1;
			}
		    }
		}
		break;

	    case file_changed: /* File has changed, upload it */
		if( prompting ) {
		    snprintf( temp, BUFSIZ, "Upload changed file: %s", current->rel_local );
		    if( !fe_ynprompt( temp ) ) continue;
		} else {
		    snprintf( temp, BUFSIZ, "Uploading changed file: %s\n", current->rel_local );
		    fe_status( temp );
		}
		if( (*the_site->driver->file_upload)( current->full_local, current->full_remote ) != PROTO_OK ) {
		    snprintf( temp, BUFSIZ, "Error! Could not upload file:\n%s\n", *(the_site->driver->last_error) );
		    fe_error( temp );
		    ret = 1;
		} else {
		    /* Successful upload */
		    current->updated = true;
		}
		break;
		
	    case file_deleted: /* File has been removed, delete it */
		if( the_site->nodelete ) break; /* unless we should keep it */
		if( prompting ) {
		    snprintf( temp, BUFSIZ, "Delete file: %s", current->rel_local );
		    if( !fe_ynprompt( temp ) ) break;
		} else {
		    snprintf( temp, BUFSIZ, "Deleting file: %s\n", current->rel_local );
		    fe_status( temp );
		}
		if( (*the_site->driver->file_delete)( current->full_remote ) != PROTO_OK ) {
		    snprintf( temp, BUFSIZ, "Error! Could not delete file:\n%s\n", *(the_site->driver->last_error) );
		    fe_error( temp );
		    ret = 1;
		} else {
		    /* Successful update - file was deleted */
		    current->updated = true;
		}
		break;

	    case file_moved:
		/* The file has been moved */
		if( prompting ) {
		    snprintf( temp, BUFSIZ, "Move file: %s -> %s", current->old->rel_local, current->rel_local );
		    if( !fe_ynprompt( temp ) ) break;
		} else {
		    snprintf( temp, BUFSIZ, "Moving file: %s -> %s\n", current->old->rel_local, current->rel_local );
		    fe_status( temp );
		}
		if( (*the_site->driver->file_move)( current->old->full_remote, 
			      current->full_remote ) != PROTO_OK ) {
		    ret = 1;
		    snprintf( temp, BUFSIZ, "Error! Could not move file:\n%s\n", *(the_site->driver->last_error) );
		    fe_error( temp );
		} else {
		    /* Successful update - file was moved */
		    current->updated = true;
		}
		break;

	    default: /* Ignore everything else */
		break;
	    }
	}
	last = current;
    }
    if( ! the_site->nodelete ) {

	/* This one must iterate through the list BACKWARDS, so
	 * directories are deleted bottom up */

	for( current=last; current!=NULL; current=current->prev ) {
	    if( current->dir && (current->diff == file_deleted ) ) {
		if( prompting ) {
		    snprintf( temp, BUFSIZ, "Delete directory: %s", current->rel_local );
		    if( !fe_ynprompt( temp ) ) continue;
		} else {
		    snprintf( temp, BUFSIZ, "Deleting directory: %s\n",current->rel_local );
		    fe_status( temp );
		}
		if( (*the_site->driver->dir_remove)( current->full_remote ) != PROTO_OK ) {
		    snprintf( temp, BUFSIZ, "Error! Could not delete directory:\n%s\n", *(the_site->driver->last_error) );
		    fe_error( temp );
		    ret = 1;
		} else {
		    /* Successful delete */
		    current->updated = true;
		}
	    }
	}
    }
 
    /* Close the connection */
    strcpy( temp, "Closing connection... " );
    (*the_site->driver->close)( );
    strcpy( temp, "done.\n" );
    site_writefiles( the_site );
    return ret;
}

/* This populates the additional names in the site_file_t struct given,
 * using information in the given site */
void site_assignnames( struct site_file_t *the_file, struct site_t *the_site ){

    the_file->full_remote = malloc( strlen(the_site->remote_root) + 
				    strlen(the_file->directory) +
				    strlen(the_file->remotename) + 1 );
    the_file->full_local = malloc( strlen(the_site->local_root) + 
				   strlen(the_file->directory) +
				   strlen(the_file->localname) + 1 );
    the_file->rel_local = malloc( strlen(the_file->directory) +
				  strlen(the_file->localname) + 2 );
    the_file->rel_remote = malloc( strlen(the_file->directory) +
				   strlen(the_file->remotename) + 2 );
    
    strcpy( the_file->full_remote, the_site->remote_root );
    strcat( the_file->full_remote, the_file->directory );
    strcat( the_file->full_remote, the_file->remotename );

    strcpy( the_file->full_local, the_site->local_root );
    strcat( the_file->full_local, the_file->directory );
    strcat( the_file->full_local, the_file->localname );

    strcpy( the_file->rel_local, "/" );
    strcat( the_file->rel_local, the_file->directory );
    strcat( the_file->rel_local, the_file->localname );

    strcpy( the_file->rel_remote, "/" );
    strcat( the_file->rel_remote, the_file->directory );
    strcat( the_file->rel_remote, the_file->remotename );

}

/* This reads off the remote files and the local files */
int site_readfiles( struct site_t *the_site ) {
    if( site_readremotefiles( the_site ) < 0 )
	return -1;
    if( site_readlocalfiles( the_site ) )
	return -2;
    /* Check for moved files - compare new files with deleted ones */
    if( the_site->checkmoved ) {
	site_checkmoved( the_site );
    }
    /* Summing up... */
    if( (the_site->numchanged +
	 the_site->nummoved + 
	 the_site->numnew +
	 (the_site->nodelete?0:the_site->numdeleted) ) > 0 ) {
	the_site->has_changed = true;
    } else {
	the_site->has_changed = false;
    }	
    return 0;
}

int site_readremotefiles( struct site_t *the_site ) {
    FILE *fp;
    char buf[BUFSIZ], /* for the line */
	tmp[BUFSIZ], /* for the current field */
	*pos, /* for the current position within buf */
	*point, /* for the curpos within tmp */
	*this_file = NULL;
    struct site_file_t *current, *previous;
    struct name_mapping_t *map;
    size_t dir_length;
    int state;
    /* Get the remote files from the storage file */
    debug( debug_full, "Reading info file: %s\n", the_site->infofile );
    fp = fopen( the_site->infofile, "r" );
    if( fp == NULL ) {
	/* This is not an error condition, since the info file WILL NOT 
	 * exist until the site has been created, which is okay, just means
	 * there are 'no' files on the remote site */
	return 1;
    }
    /* The file exists, so read it.
     * Format: one item / line, tab-separated fields.
     * First field is filename of item.
     * Second field is 'dir' for directory items, or mtime for files
     * Third field (files only) is size of file */
    current = NULL;
    while( fgets( buf, BUFSIZ, fp) != NULL ) {
	/* Create a new file item and lob it on the end of the linked 
	 * list */
	previous = current;
	current = malloc( sizeof(struct site_file_t) );
	memset( current, 0, sizeof(struct site_file_t) );
	if( previous!=NULL ) {
	    previous->next = current;
	} else {
	    the_site->files = current;
	}
	current->prev = previous;
	/* Now parse the line. Simple automaton, states are:
	 *  0: Reading filename, field 1
	 *  1: Reading date/time stamp or DIRWORD, field 2
	 *  2: Reading file size, field 3 (if file)
	 *  3: Junk state - consume-all-crud-till-end-of-line.
	 * 
	 * We read the current field into tmp char by char.
	 * point is the current point within tmp. */
	state = 0;
	point = tmp;
	for( pos = buf; *pos!='\0'; pos++ ) {
	    switch( state ) {
	    case 0:
		if( *pos == '\t' ) {
		    /* End of the filename */
		    *point = '\0';
		    this_file = strdup( tmp );
		    point = tmp;
		    state = 1;
		} else {
		    /* Still in the filename */
		    *(point++) = *pos;
		}
		break;
	    case 1:
		if( *pos == '\t' || *pos == '\n' ) {
		    /* End of the second field */
		    *point = '\0';
		    if( strlen( tmp ) > 0 ) {
			if( strcmp( tmp, DIRWORD ) == 0 ) {
			    /* It's a directory! */
			    current->dir = true;
			    state = 3; /* that's all we need */
			} else {
			    /* It's a file! - field 2 is the mtime */
			    current->dir = false;
			    current->remotetime = atol( tmp );
			    point = tmp;
			    state = 2;
			}
		    } else {
			/* Corrupt line, we need at least two fields */
			debug( 0, "Warning: Corrupt info file line, skipped.\n" );
			state = 3;
		    }			
		} else {
		    /* Within the second field */
		    *(point++) = *pos;
		}
		break;
	    case 2:
		if( *pos == '\n' ) {
		    /* End of the size field */
		    *point = '\0';
		    current->remotesize = atol( tmp );
		    state = 3;
		} else {
		    /* Within the file size field */
		    *(point++) = *pos;
		}
		break;
	    case 3:
		break;
	    }
	}
	if( this_file==NULL ) {
	    debug(0, "Warning: corrupt line in info file, skipped.\n" );
	    continue;
	}
	current->diff = file_deleted;
	current->remotename = basename( this_file );
	/* we are going to chop the leading / of this_file */
	dir_length = strlen(this_file) - strlen(current->remotename ) - 1;
	debug( debug_total, "dir_length: %d\n", dir_length );
	current->directory = malloc( dir_length + 1 );
	strncpy( current->directory, this_file + 1, dir_length );
	*(current->directory+dir_length) = '\0'; /* null-term */
	debug( debug_total, "Split %s is: Dir [%s], name [%s]\n",
	       this_file, current->directory, current->remotename );
	the_site->numdeleted++;
	if( current->dir ) {
	    /* It's a dir */
	    current->localname = current->remotename;
	} else {
	    /* It's a file, check if it is mapped to a different
	     * local file name */
	    for( map=the_site->renames; map!=NULL; map=map->next )
		if( strcmp( map->remote, current->remotename ) == 0 )
		    break;
	    if( map!=NULL ) {
		/* It's mapped */
		current->localname = strdup( map->local );
	    } else {
		current->localname = strdup( current->remotename );
	    }
	}
	site_assignnames( current, the_site );
	debug( debug_total, "Remote: [%s] - file: size: %ld  time: %ld\n",
	       current->remotename, current->remotesize, current->remotetime );
	debug( debug_total, 
	       "Directory: [%s]  Local: %s   Remote: %s\n"
	       "rel_local: %s  rel_remote: %s\n"
	       "full_local: %s  full_remote: %s\n",
	       current->directory, current->localname,
	       current->remotename, current->rel_local,
	       current->rel_remote, current->full_local,
	       current->full_remote );
    }
    fclose( fp );
    return 0;
}

/* Read the local site files... 
 * A stack is used for directories within the site - this is not recursive.
 * Each item on the stack is a FULL PATH to the directory, i.e., including
 * the local site root. */

int site_readlocalfiles( struct site_t *the_site ) {
    char *dirstack[MAXDIRS], *this, *full, *fname;
    struct stat item;
    DIR *curdir;
    struct dirent *ent;
    int dirtop = 0, excl;
    size_t dir_length;
    struct name_mapping_t *map;
    struct site_file_t *current, *last;

    /* Push the root directory on to the stack */
    dirstack[dirtop++] = the_site->local_root;
    
    /* Now, for all items in the stack, process all the files, and
     * add the dirs to the stack. */

    while( dirtop > 0 ) {
	/* Pop the top item */
	this = dirstack[--dirtop];
	debug( debug_total, "Dir: %s\n", this );
	curdir = opendir( this );
	if( curdir == NULL ) {
	    continue;
	}

	/* Now read all the directory entries */

	while( (ent = readdir( curdir )) != NULL ) {
	    
	    full = malloc( strlen( this ) + strlen( ent->d_name ) + 1 );
	    strcpy( full, this );
	    strcat( full, ent->d_name );
	    debug( debug_total, "Item: %s - ", ent->d_name );
	    if( lstat( full, &item ) == -1 ) {
		perror( "stat" );
		continue;
	    }
	    
	    /* Is this a symlink? */
	    if( S_ISLNK( item.st_mode ) ) {
		debug( debug_total, "symlink - " );
		if( the_site->symlinks == sitesym_ignore ) {
		    /* Just skip it */
		    debug( debug_total, "ignoring.\n" );
		    continue;
		}
		debug( debug_total, "followed - " );
		/* Else, carry on as normal, stat the real file */
		if( stat( full, &item ) == -1 ) {
		    /* It's probably a broken link */
		    debug( debug_total, "broken.\n" );
		    continue;
		}
	    }

	    /* Check for excludes */
	    if( the_site->numexcludes > 0 ) {
		debug( debug_total, "\nChecking excludes:\n" );
		for( excl=0; excl != the_site->numexcludes; excl++ ) {
		    debug( debug_total, "%s ", the_site->excludes[excl] );
		    if( fnmatch( the_site->excludes[excl], ent->d_name, 0 ) == 0 )
			break;
		}
		if( excl < the_site->numexcludes ) { /* excluded */
		    debug( debug_total, "- matched\n" );
		    continue;
		}
		debug( debug_total, "... not matched.\n" );
	    }
	    
	    /* Now process it */
	    last = NULL;
	    /* This is the rel_local of this file - i.e., everything
	     * apart from the local root */
	    fname = (char *)full+strlen(the_site->local_root)-1;
   
	    for( current=the_site->files; current!=NULL; current=current->next ) {
		if( strcmp( current->rel_local, fname ) == 0 )
		    /* Match found! */
		    break;
		last = current;
	    }
	    
	    if( S_ISREG( item.st_mode ) ) {
		debug( debug_total, "file - " );
		if( current == NULL ) {
		    /* This is a new local file... */
		    debug( debug_total, "new - " );
		    the_site->numnew++;
		    current = malloc( sizeof(struct site_file_t) );
		    memset( current, 0, sizeof(struct site_file_t) );
		    if( (current->next = the_site->files) != NULL )
			current->next->prev = current;
		    the_site->files = current;
		    current->localname = strdup( basename( fname ) );
		    /* This is the length of the directory name...
		     * fname = "/full/file/name.ext", localname = "name.ext"
		     * we want directory = "full/file/"... do the maths */
		    dir_length = strlen(fname) - strlen(current->localname)-1;
		    debug( debug_total, "dir_length = %d\n", dir_length );
		    current->directory = malloc( dir_length + 1 );
		    strncpy( current->directory, fname+1, dir_length );
		    *(current->directory+dir_length) = '\0'; /* nullterm */
		    debug( debug_total, "directory = %s\n", current->directory );
		    /* Check if it's mapped to a different remote name */
		    for( map = the_site->renames; map!=NULL; map=map->next )
			if( strcmp( map->local, current->localname ) == 0 )
			    break;
		    /* now assign the remote name appropriately */
		    if( map!=NULL ) {
			debug( debug_total, "Map: local %s -> %s\n",
			       current->rel_local, current->remotename );
			current->remotename = strdup( map->remote );
		    } else {
			/* it's not mapped */
			current->remotename = strdup( current->localname );
		    }
		    debug( debug_total, "Split: %s - dir [%s], localname: [%s]\n", fname, current->directory, current->localname );
		    site_assignnames( current, the_site );
		    current->localtime = item.st_mtime;
		    current->localsize = item.st_size;
		    current->diff = file_new;
		    current->mode = item.st_mode;
		    current->dir = false;
/*		    debug( debug_total, "new; size: %ld, time: %ld\n", item.st_size, item.st_mtime );  */
		} else {
		    /* Match found... compare it */
		    current->localtime = item.st_mtime;
		    current->localsize = item.st_size;
		    the_site->numdeleted--;
		    if( current->localtime > current->remotetime ) {
			current->diff = file_changed;
			the_site->numchanged++;
			current->mode = item.st_mode;
			debug( debug_total, "changed (%ld -> %ld)\n", current->localtime, current->remotetime );
		    } else {
			current->diff = file_unchanged;
			current->updated = true;
			the_site->numunchanged++;
			debug( debug_total, "unchanged.\n" );
		    }
		    /* The names will already have been assigned */
		}
	    } else if( S_ISDIR( item.st_mode ) ) {
		/* THIS IS A DIRECTORY! */
		if( strcmp( ent->d_name, "." )==0 ||
		    strcmp( ent->d_name, ".." )==0 ) {
		    debug( debug_total, "ignored.\n" );
		    continue;
		}
		
		debug( debug_total, "directory - " );
		if( dirtop < MAXDIRS ) {
		    dirstack[dirtop] = malloc( strlen( full ) + 2 );
		    strcpy( dirstack[dirtop], full );
		    strcat( dirstack[dirtop++], "/" );
		} else {
		    /* No more room in stack */
		    debug( 0, 
"Warning: directory search depth exceeded. Increase MAXDIRS in sites.c.\n\
Directory %s is being skipped.\n", full );
		}
		if( current == NULL ) {
		    /* New Directory */
		    the_site->numnew++;
		    current = malloc( sizeof(struct site_file_t) );
		    memset( current, 0, sizeof(struct site_file_t) );
		    /* Directories go on the END of the list */
		    if( last ) {
			last->next = current;
		    } else {
			the_site->files = current;
		    }
		    current->prev = last;
		    current->localname = strdup( basename( fname ) );
		    current->remotename = strdup( current->localname );
		    /* Work out the directory name */
		    dir_length = strlen(fname) - strlen(current->localname )-1;
		    current->directory = malloc( dir_length + 1 );
		    strncpy( current->directory, fname+1, dir_length );
		    *(current->directory+dir_length) = '\0'; /* null-term */
		    debug( debug_total, "Split: %s - dir [%s], localname: [%s]\n", fname, current->directory, current->localname );
		    site_assignnames( current, the_site );
		    current->dir = true;
		    current->diff = file_new;
		    current->mode = item.st_mode;
		    current->localsize = item.st_size;
		    debug( debug_total, "new.\n" ); 
		} else {
		    /* It's an existing directory */
		    the_site->numdeleted--;
		    the_site->numunchanged++;
		    current->diff = file_unchanged;
		    current->updated = true;
		    debug( debug_total, "existing.\n" ); 
		}
	    } else {
		debug( debug_total, "something else.\n" );
	    }
	    debug( debug_total, 
		   "Directory: [%s]  Local: %s   Remote: %s\n"
		   "rel_local: %s  rel_remote: %s\n"
		   "full_local: %s  full_remote: %s\n",
		   current->directory, current->localname,
		   current->remotename, current->rel_local,
		   current->rel_remote, current->full_local,
		   current->full_remote );
	}
	/* Close the open directory */
	closedir( curdir );
    }
    return 0;
}

void site_checkmoved( struct site_t *the_site ) {
    /* Have any files been moved?
     * We look for new files, then see if we can find a deleted file
     * which matches filename, mod time, and size exactly.
     * 
     * This is simplistic, and it is possible to get it wrong in some
     * (unusual) circumstances:
     *    Old: /bar/file1.zip, /foo/file1.zip 
     *    New: /new/file1.zip.
     * If all the file1.zip's have the same size and mod time, it
     * IS NOT POSSIBLE to know which one has become /new/file1.zip without 
     * storing remembering the contents of /bar/file1.zip somehow and 
     * comparing, or using a filesystem which remembers stuff. In any case
     * it's pretty silly to try - so the whole shabang is optional.
     *
     * It would probably be better to search for DELETED files, then 
     * check them against NEW files, since files are more often added
     * to sites then they are deleted. But hey, this is C, who'll notice?
     */
    struct site_file_t *search, *current;
    /*  char *bname, *tmp; */
    /* First off, find the new files... */
    debug( debug_full, "Checking for moved files...\n" );
    for( search=the_site->files; search!=NULL; search=search->next ) {
	if( search->diff != file_new )
	    continue;
	/* OK, we have us a new file... now iterate through the deleted
	 * files and look for a match */
	debug( debug_full, "- New file: %s\nComparing:", search->rel_local );
	for( current=the_site->files; current!=NULL ; current=current->next ) {
	    if( (current->diff != file_deleted) || current->dir ) {
		/* Ignore files which aren't deleted, and dirs, completely.
		 * (search->diff==file_new, so we DON'T need to check
		 * that current != search as well) */
		continue;
	    }
	    debug( debug_full, " %s", current->rel_local );
	    if( (strcmp( search->localname, current->localname ) == 0) &&
		current->remotesize == search->localsize &&
		current->remotetime == search->localtime ) {
		debug( debug_full, " - matched!\n" );
		break;
	    }
	}
	if( current != NULL ) {
	    /* OK, current == the deleted file, search == the new file.
	     * First make sure search knows where it WAS before the move... */
	    search->old = current;
	    /* Remove the deleted file from the list so it doesn't
	     * actually get deleted by site_update */
	    LL_DELETE( current, the_site->files );
	    /* Diddle the totals a bit */
	    search->diff = file_moved;
	    the_site->numdeleted--;
	    the_site->numnew--;
	    the_site->nummoved++;
	} else {
	    debug( debug_full, " - No match found.\n" );
	}
    }
    debug( debug_full, "Finished checking for moved files.\n" );
}

/* This writes the remote files list back to the file
 * and has some hairy logic in the middle, but is otherwise
 * pretty simple */

#define WRITE_REMOTE_FILE( fp, f ) \
fprintf( fp, "%s\t%ld\t%ld\n", f->rel_remote, f->remotetime, f->remotesize );
#define WRITE_LOCAL_FILE( fp, f  ) \
fprintf( fp, "%s\t%ld\t%ld\n", f->rel_local, f->localtime, f->localsize );

int site_writefiles( struct site_t *the_site ) {
    FILE *fp;
    struct site_file_t *current;
    fp = fopen( the_site->infofile, "w" );
    if( fp == NULL ) {
	/* Couldn't open! */
	perror( "fopen" );
	fprintf( stderr, "%s: Could not open site info file for writing: %s\n", progname, the_site->infofile );
	return 1;
    }
    for( current = the_site->files; current!=NULL; current=current->next ) {
	/* This is the 20-line version of the 60-line monster below.
	 * The 60-line monster is, on the other hand, comprehensible.
	 * This version might give a fractional speed increase
	if( current->dir ) {
	    if( (current->diff==file_unchanged) ||
		(current->updated && current->diff==file_new ) ||
		(!current->updated && current->diff==file_deleted ) ) {
		... print directory is there ...
	    } else {
		... print nothing ...
	    }
	} else { 
	    if( current->diff==file_unchanged ||
		(!current->updated && (current->file==file_changed ||
				       current->file==file_deleted) ) ) {
		... print remote file ...
	    } else if( current->updated &&
		       (current->diff==file_changed ||
			current->diff==file_moved ||
			current->diff==file_new) ) {
		... print local file ...
	    } else if( !current->updated && current->diff==file_moved ) {
		... print OLD file ...
	    } else {
		... print nothing ...
	    }
	} */
	   
	if( current->dir ) {
	    if( current->updated ) {
		switch( current->diff ) {
		case file_new:
		case file_unchanged:
		    /* It's new, or it's already there... */
		    fprintf( fp, "%s\t%s\n", current->rel_local, DIRWORD );
		    break;
		case file_deleted:
		case file_moved:
		case file_changed:
		default:
		    /* It's not there */
		    break;
		} 
	    } else {
		/* The update was not successful */
		switch( current->diff ) {
		case file_unchanged: /* ... it was there anyway */
		case file_deleted: /* ... but it wasn't deleted */
		    /* So the directory exists remotely */
		    fprintf( fp, "%s\t%s\n", current->rel_remote, DIRWORD );
		    break;
		case file_new: /* ... but it wasn't created */
		default:
		    /* So the directory does not exist remotely */
		    break;
		}
	    }
	} else {
	    if( current->updated ) {
		switch( current->diff ) {
		case file_unchanged:
		    /* Write state of remote file */
		    WRITE_REMOTE_FILE( fp, current );
		    break;
		case file_moved:
		case file_changed:
		case file_new:
		    /* Write state of local file */
		    WRITE_LOCAL_FILE( fp, current );
		    break;
		case file_deleted:
		    /* So don't write anything out */
		    break;
		}
	    } else {
		/* Not updated */
		switch( current->diff ) {
		case file_changed: /* ... but it hasn't been updated */
		case file_unchanged: 
		case file_deleted: /* ... but it hasn't been deleted */
		    /* Write out the remote time */
		    WRITE_REMOTE_FILE( fp, current );
		    break;
		case file_moved: 
		    /* But it hasn't actually been moved, so write out
		     * the 'deleted' file (where the moved file WAS), and 
		     * ignore the 'new' file (where it was moved TO) */
		    WRITE_REMOTE_FILE( fp, current->old );
		    break;
		case file_new: /* So ignore it... */
		    break;
		}
	    }
	}
    }
    fclose( fp );
    return 0;
}

#undef WRITE_REMOTE_FILE
#undef WRITE_LOCAL_FILE

/* Simply marks all the files in the given site as 'updated'. */
int site_catchup( struct site_t *the_site ) {
    struct site_file_t *current;
    for( current=the_site->files; current!=NULL; current=current->next ) {
	current->updated = true;
    }
    return 0;
}

int site_initialize( struct site_t *the_site ) {
    /* For this, we have to mark all files as 'new' locally, and 
     * NOT updated. So, we pretend there is NOTHING on the remote site */
    struct site_file_t *current;
    
    for( current=the_site->files; current!=NULL; current=current->next ) {
	current->updated = false;
	current->diff = file_new;
    }
    /* And that's it. I think */
    return 0;
}
